/** 
 * <a href="http://servicemix.org">ServiceMix: The open source ESB</a> 
 * 
 * Copyright 2005 RAJD Consultancy Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.servicemix.jbi.framework;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Properties;
import java.util.Set;
import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.management.LifeCycleMBean;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.management.ObjectName;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.servicemix.jbi.container.ActivationSpec;
import org.servicemix.jbi.messaging.DeliveryChannelImpl;

/**
 * ComponentConnector is used internally for message routing
 * 
 * @version $Revision: 1.5 $
 */
public class LocalComponentConnector extends ComponentConnector {
    private static final Log log = LogFactory.getLog(LocalComponentConnector.class);
    private Component component;
    private ComponentContextImpl context;
    private ActivationSpec activationSpec;
    private DeliveryChannelImpl deliveryChannel;
    private ObjectName mbeanName;
    private ObjectName extendedMBeanName;
    private ComponentMBean componentMBean;
    private boolean pojo;

    /**
     * Default Constructor
     */
    public LocalComponentConnector() {
        super();
    }

    /**
     * Construct with it's id and delivery channel Id
     * 
     * @param name
     * @param description
     * @param component
     * @param dc
     * @param binding
     * @param service
     */
    public LocalComponentConnector(ComponentNameSpace name, String description, Component component,
            DeliveryChannelImpl dc, boolean binding, boolean service) {
        super(name);
        this.component = component;
        this.deliveryChannel = dc;
        packet.setDescription(description);
        packet.setBinding(binding);
        packet.setService(service);
    }

    /**
     * @return true if the Component is local to the Container
     */
    public boolean isLocal() {
        return true;
    }

    /**
     * Get the Context
     * 
     * @return the Context
     */
    public ComponentContextImpl getContext() {
        return context;
    }

    /**
     * Set the Context
     * 
     * @param context
     */
    public void setContext(ComponentContextImpl context) {
        this.context = context;
    }

    /**
     * Add an activated endpoint
     * 
     * @param endpoint
     */
    public void addActiveEndpoint(ServiceEndpoint endpoint) {
        packet.addActiveEndpoint(endpoint);
    }

    /**
     * remove an activated endpoint
     * 
     * @param endpoint
     */
    public void removeActiveEndpoint(ServiceEndpoint endpoint) {
        packet.removeActiveEndpoint(endpoint);
    }

    /**
     * Add an external activated endpoint
     * 
     * @param endpoint
     */
    public void addExternalActiveEndpoint(ServiceEndpoint endpoint) {
        packet.addExternalActiveEndpoint(endpoint);
    }

    /**
     * remove an external activated endpoint
     * 
     * @param endpoint
     */
    public void removeExternalActiveEndpoint(ServiceEndpoint endpoint) {
        packet.removeExternalActiveEndpoint(endpoint);
    }

    /**
     * Get the Set of external activated endpoints
     * 
     * @return the activated endpoint Set
     */
    public Set getExternalActiveEndpoints() {
        return packet.getExternalActiveEndpoints();
    }

    /**
     * @return Returns the component.
     */
    public Component getComponent() {
        return component;
    }

    /**
     * @param component The component to set.
     */
    public void setComponent(Component component) {
        this.component = component;
    }

    /**
     * @return Returns the deliveryChannel.
     */
    public DeliveryChannelImpl getDeliveryChannel() {
        return deliveryChannel;
    }

    /**
     * @param deliveryChannel The deliveryChannel to set.
     */
    public void setDeliveryChannel(DeliveryChannelImpl deliveryChannel) {
        this.deliveryChannel = deliveryChannel;
    }

    /**
     * @return the ActivateionSpec
     */
    public ActivationSpec getActivationSpec() {
        return activationSpec;
    }

    /**
     * Set the ActivationSpec
     * 
     * @param activationSpec
     */
    public void setActivationSpec(ActivationSpec activationSpec) {
        this.activationSpec = activationSpec;
    }

    /**
     * @return Returns the mbeanName.
     */
    public ObjectName getMbeanName() {
        return mbeanName;
    }

    /**
     * @return Returns the ComponentMBean
     */
    public ComponentMBean getComponentMBean() {
        return componentMBean;
    }

    /**
     * @param mbeanName The mbeanName to set.
     * @param mbean
     */
    public void setMBean(ObjectName mbeanName, ComponentMBean mbean) {
        this.mbeanName = mbeanName;
        this.componentMBean = mbean;
    }

    /**
     * @return Returns the extendedMBeanName.
     */
    public ObjectName getExtendedMBeanName() {
        return extendedMBeanName;
    }

    /**
     * @param extendedMBeanName The extendedMBeanName to set.
     */
    public void setExtendedMBeanName(ObjectName extendedMBeanName) {
        this.extendedMBeanName = extendedMBeanName;
    }

    /**
     * @return Returns the pojo.
     */
    public boolean isPojo() {
        return pojo;
    }

    /**
     * @param pojo The pojo to set.
     */
    public void setPojo(boolean pojo) {
        this.pojo = pojo;
    }

    /**
     * write the current running state of the Component to disk
     */
    void writeRunningState() {
        String componentName = getComponentNameSpace().getName();
        if (componentMBean != null) {
            try {
                String currentState = componentMBean.getCurrentState();
                File stateFile = context.getContainer().getEnvironmentContext().getComponentStateFile(componentName);
                FileOutputStream out = new FileOutputStream(stateFile);
                Properties props = new Properties();
                props.setProperty("state", currentState);
                props.store(out, "Running state for Component: " + componentName);
            }
            catch (IOException e) {
                log.error("Failed to write current running state for Component: " + componentName, e);
            }
        }
        else {
            log.warn("No componentMBean available for Component: " + componentName);
        }
    }

    /**
     * Read the last running state from disk and set the component to this state
     * 
     * @throws IOException
     * @throws JBIException
     */
    void setRunningState() throws JBIException, IOException {
        String componentName = getComponentNameSpace().getName();
        String runningState = getRunningState();
        log.info("Setting running state for Component: " + componentName + " to " + runningState);
        if (runningState != null && componentMBean != null) {
            if (runningState.equals(LifeCycleMBean.RUNNING)) {
                componentMBean.start();
            }
            if (runningState.equals(LifeCycleMBean.STOPPED)) {
                componentMBean.stop();
            }
        }
    }

    /**
     * @return the current running state from disk
     */
    String getRunningState() {
        String result = LifeCycleMBean.UNKNOWN;
        if (componentMBean != null) {
            String componentName = getComponentNameSpace().getName();
            File stateFile = context.getContainer().getEnvironmentContext().getComponentStateFile(componentName);
            try {
                FileInputStream in = new FileInputStream(stateFile);
                Properties props = new Properties();
                props.load(in);
                result = props.getProperty("state", result);
            }
            catch (IOException e) {
                log.error("Failed to write current running state for Component: " + componentName, e);
            }
        }
        return result;
    }
}
