/** 
 * <a href="http://servicemix.org">ServiceMix: The open source ESB</a> 
 * 
 * Copyright 2005 RAJD Consultancy Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/

package org.servicemix.jbi.nmr;

import javax.jbi.JBIException;
import javax.jbi.component.Component;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.resource.spi.work.WorkManager;
import javax.xml.namespace.QName;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.servicemix.jbi.container.ActivationSpec;
import org.servicemix.jbi.container.JBIContainer;
import org.servicemix.jbi.framework.ComponentConnector;
import org.servicemix.jbi.framework.ComponentContextImpl;
import org.servicemix.jbi.framework.ComponentNameSpace;
import org.servicemix.jbi.framework.LocalComponentConnector;
import org.servicemix.jbi.framework.Registry;
import org.servicemix.jbi.management.BaseLifeCycle;
import org.servicemix.jbi.management.ManagementContext;
import org.servicemix.jbi.messaging.ExchangePacket;
import org.servicemix.jbi.messaging.MessageExchangeImpl;
import org.servicemix.jbi.nmr.flow.Flow;
import org.servicemix.jbi.nmr.flow.FlowProvider;
import org.servicemix.jbi.resolver.ConsumerComponentEndpointFilter;
import org.servicemix.jbi.resolver.EndpointChooser;
import org.servicemix.jbi.resolver.EndpointFilter;
import org.servicemix.jbi.resolver.EndpointResolver;
import org.servicemix.jbi.resolver.FirstChoicePolicy;
import org.servicemix.jbi.resolver.ProducerComponentEndpointFilter;
import org.servicemix.jbi.servicedesc.ServiceEndpointImpl;

/**
 * The Broker handles Nomalised Message Routing within ServiceMix
 *
 * @version $Revision: 1.18 $
 */
public class Broker  extends BaseLifeCycle {
    private JBIContainer container;
    private Registry registry;
    private String flowName = "seda";
    private WorkManager workManager;
    private Flow flow;
    private final static Log log = LogFactory.getLog(Broker.class);
    private EndpointChooser defaultServiceChooser = new FirstChoicePolicy();
    private EndpointChooser defaultInterfaceChooser = new FirstChoicePolicy();

    /**
     * Constructor
     */
    public Broker() {
    }
    
    /**
     * Get the description
     * @return description
     */
    public String getDescription(){
        return "Normalized Message Router";
    }
    
    /**
     * @return Returns the workManager.
     */
    public WorkManager getWorkManager() {
        return workManager;
    }

    /**
     * @param workManager The workManager to set.
     */
    public void setWorkManager(WorkManager workManager) {
        this.workManager = workManager;
    }    

    /**
     * initialize the broker
     * @param container
     * @throws JBIException
     */
    public void init(JBIContainer container) throws JBIException {
        this.container = container;
        this.workManager = container.getWorkManager();
        this.registry = container.getRegistry();
        if (this.flow == null) {
            this.flow = FlowProvider.getFlow(flowName);
        }
        this.flow.init(this);
    }
    
    /**
     * Get the name of the Container
     * @return containerName
     */
    public String getContainerName(){
        return container.getName();
    }
    
    /**
     * Get the ManagementContext
     * @return the managementContext
     */
    public ManagementContext getManagementContext(){
        return container.getManagementContext();
    }
    
    /**
     * Get the Registry
     * @return the registry
     */
    public Registry getRegistry(){
        return registry;
    }
    /**
     * start brokering
     * @throws JBIException
     */
    public void start() throws JBIException{
        flow.start();
        super.start();
    }
    
    /**
     * stop brokering
     * @throws JBIException
     */
    public void stop() throws JBIException{
        flow.stop();
        super.stop();
    }
    
    /**
     * shutdown all Components
     * @throws JBIException
     */
    public void shutDown() throws JBIException{
        stop();
        flow.shutDown();
        super.shutDown();
    }
    
    /**
     * @return Returns the flow.
     */
    public String getFlowName() {
        return flowName;
    }
    /**
     * @param flowName The flow to set.
     */
    public void setFlowName(String flowName) {
        this.flowName = flowName;
    }
    
    /**
     * Set the flow
     * @param flow
     */
    public void setFlow(Flow flow){
        this.flow = flow;
    }
    
    /**
     * @return the Flow
     */
    public Flow getFlow(){
        return this.flow;
    }

    /**
     * Route an ExchangePacket to a destination
     *
     * @param exchange
     * @throws JBIException
     */
    public void sendExchangePacket(MessageExchangeImpl exchange) throws JBIException {
        ExchangePacket packet = exchange.getPacket();
        if (packet.isOutbound()) {
            ComponentNameSpace destinationId = null;
            ServiceEndpointImpl theEndpoint = packet.getEndpoint();
            QName serviceName = packet.getServiceName();
            QName interfaceName = packet.getInterfaceName();

            // get the context which created the exchange
            ComponentContextImpl context = exchange.getSourceContext();
            boolean remoteDestination = false;
            if (theEndpoint == null) {
                //check in order, ServiceName then InterfaceName
                //if remote, send on for remote broker to do the routing
                //check to see if there is a match on the serviceName
                if (serviceName != null) {
                    ServiceEndpoint[] endpoints = registry.getEndpointsForService(serviceName);
                    //check against name (if set)
                    //theEndpoint = matchEndpointByName(endpoints,endpointName);
                    theEndpoint = (ServiceEndpointImpl) getServiceChooser(exchange).chooseEndpoint(endpoints, context, exchange);

                    // TODO should we match against operationName / interfaceName?

                    if (theEndpoint == null) {
                        //check to see if the target component is remote
                        ComponentConnector consumerConnector = registry.getComponentConnectorByExternalServiceName(serviceName);
                        if (consumerConnector != null) {
                            if (consumerConnector instanceof LocalComponentConnector) {
                                theEndpoint = registry.getMatch((LocalComponentConnector) consumerConnector, packet
                                        .getSourceId(), exchange, endpoints);
                            }
                            else {
                                remoteDestination = true;
                                destinationId = consumerConnector.getComponentPacket().getComponentNameSpace();
                            }
                        }
                        else {
                            log.warn("SericeName (" + serviceName
                                    + ") specified for routing, but can't find it registered");
                        }
                    }
                }
                if (theEndpoint == null && !remoteDestination && interfaceName != null) {
                    ServiceEndpoint[] endpoints = registry.getEndpoints(interfaceName);
                    //check against name (if set)
                    //theEndpoint = matchEndpointByName(endpoints,endpointName);
                    theEndpoint = (ServiceEndpointImpl) getInterfaceChooser(exchange).chooseEndpoint(endpoints, context, exchange);
                    if (theEndpoint == null) {
                        //check to see if the target component is remote
                        ComponentConnector consumerConnector = registry.getComponentConnector(interfaceName);
                        if (consumerConnector != null) {
                            if (consumerConnector instanceof LocalComponentConnector) {
                                theEndpoint = registry.getMatch((LocalComponentConnector) consumerConnector, packet
                                        .getSourceId(), exchange, endpoints);
                            }
                            else {
                                remoteDestination = true;
                                destinationId = consumerConnector.getComponentPacket().getComponentNameSpace();
                            }
                        }
                        else {
                            log.warn("SericeName (" + serviceName
                                    + ") specified for routing, but can't find it registered");
                        }
                    }
                }
                if (theEndpoint == null) {
                    // lets use the resolver on the activation spec if applicable

                    ActivationSpec activationSpec = exchange.getActivationSpec();
                    if (activationSpec != null) {
                        EndpointResolver destinationResolver = activationSpec.getDestinationResolver();
                        if (destinationResolver != null) {
                            try {
                                EndpointFilter filter = createEndpointFilter(context, exchange);
                                theEndpoint = (ServiceEndpointImpl) destinationResolver.resolveEndpoint(context, exchange, filter);
                            }
                            catch (JBIException e) {
                                throw new MessagingException("Failed to resolve endpoint: " + e, e);
                            }
                        }
                    }
                }
            }
            if (theEndpoint != null || remoteDestination) {
                packet.setEndpoint(theEndpoint);
                if (destinationId == null) {
                    destinationId = packet.getEndpoint().getComponentNameSpace();
                }
                packet.setDestinationId(destinationId);
            }
            else {
                throw new MessagingException("Could not find route for exchange: " + exchange
                        + " for service: " + serviceName + " and interface: " + interfaceName);
            }
           if (log.isTraceEnabled()){
               log.trace("routing exchange " + packet + " TO: " + theEndpoint);
           }
        }
        
        flow.send(packet);
    }

    /**
     * @return the default EndpointChooser
     */
    public EndpointChooser getDefaultInterfaceChooser() {
        return defaultInterfaceChooser;
    }

    /**
     * Set the default EndpointChooser
     * @param defaultInterfaceChooser
     */
    public void setDefaultInterfaceChooser(EndpointChooser defaultInterfaceChooser) {
        this.defaultInterfaceChooser = defaultInterfaceChooser;
    }

    /**
     * @return the default EndpointChooser
     */
    public EndpointChooser getDefaultServiceChooser() {
        return defaultServiceChooser;
    }

    /**
     * Set default EndpointChooser
     * @param defaultServiceChooser
     */
    public void setDefaultServiceChooser(EndpointChooser defaultServiceChooser) {
        this.defaultServiceChooser = defaultServiceChooser;
    }

    /**
     * Returns the endpoint chooser for endpoints found by service which will use the chooser on the exchange's
     * activation spec if available otherwise will use the default
     * @param exchange
     * @return the EndpointChooser
     */
    protected EndpointChooser getServiceChooser(MessageExchangeImpl exchange) {
        EndpointChooser chooser = null;
        ActivationSpec activationSpec = exchange.getActivationSpec();
        if (activationSpec != null) {
            chooser = activationSpec.getServiceChooser();
        }
        if (chooser == null) {
            chooser = defaultServiceChooser;
        }
        return chooser;
    }


    /**
     * Returns the endpoint chooser for endpoints found by service which will use the chooser on the exchange's
     * activation spec if available otherwise will use the default
     * @param exchange
     * @return the EndpointChooser
     */
    protected EndpointChooser getInterfaceChooser(MessageExchangeImpl exchange) {
        EndpointChooser chooser = null;
        ActivationSpec activationSpec = exchange.getActivationSpec();
        if (activationSpec != null) {
            chooser = activationSpec.getInterfaceChooser();
        }
        if (chooser == null) {
            chooser = defaultInterfaceChooser;
        }
        return chooser;
    }

    /**
     * Factory method to create an endpoint filter for the given component context and message exchange
     * @param context
     * @param exchange
     * @return the EndpointFilter
     */
    protected EndpointFilter createEndpointFilter(ComponentContextImpl context, MessageExchangeImpl exchange) {
        Component component = context.getComponent();
        ExchangePacket packet = exchange.getPacket();
        if (packet.isOutbound()) {
            return new ConsumerComponentEndpointFilter(component);
        }
        else {
            return new ProducerComponentEndpointFilter(component);
        }
    }

    
}